from bf_gen import BFBuilder

def test_builder_branch():
    builder = BFBuilder()

    builder.move(0)
    builder.init_with_num(1)
    with builder.branch():
        builder.move(1)
        builder.init_with_letter("A")
        builder.output()

    builder.move(0)
    builder.init_with_zero()
    with builder.branch():
        builder.move(1)
        builder.init_with_letter("B")
        builder.output()

    assert builder.interpret() == "A"

def test_builder_hello_world():
    builder = BFBuilder()

    for letter in "Hello World":
        builder.init_with_letter(letter)
        builder.output()

    assert builder.interpret() == "Hello World"

def test_builder_triangle():
    builder = BFBuilder()
    triangle_size = 5

    DESCENDING = 0
    ASCENDING = 1
    SYMBOL = 2
    TEMP = 3

    builder.move(DESCENDING)
    builder.init_with_num(triangle_size)

    with builder.loop():
        builder.add(-1)

        builder.move(TEMP)
        builder.init_with_zero()

        builder.move(ASCENDING)
        builder.add(1)
        builder.copy(SYMBOL, TEMP)

        builder.move(SYMBOL)

        with builder.loop():
            builder.add(-1)

            builder.move(TEMP)
            builder.init_with_letter("#")
            builder.output()

        builder.move(TEMP)
        builder.init_with_letter("\n")
        builder.output()

    assert builder.interpret() == "#\n##\n###\n####\n#####\n"



class BFPreference:
    """
    Stores language preference
    """
    def __init__(self):
        self.increment = "+"
        self.decrement = "-"
        self.move_right = ">"
        self.move_left = "<"
        self.input = ","
        self.output = "."
        self.start_loop = "["
        self.end_loop = "]"

    def brainf__k(self):
        """
        Have BFBuilder use Brainf**k.

        What is Brainf**k : https://esolangs.org/wiki/Brainfuck
        """
        self.increment = "+"
        self.decrement = "-"
        self.move_right = ">"
        self.move_left = "<"
        self.input = ","
        self.output = "."
        self.start_loop = "["
        self.end_loop = "]"

    def ook(self):
        """
        Have BFBuilder use Ook!

        What is Ook! : https://esolangs.org/wiki/Ook!
        """
        self.increment = "Ook.Ook."
        self.decrement = "Ook!Ook!"
        self.move_right = "Ook.Ook?"
        self.move_left = "Ook?Ook."
        self.input = "Ook.Ook!"
        self.output = "Ook!Ook."
        self.start_loop = "Ook!Ook?"
        self.end_loop = "Ook?Ook!"

    def blub(self):
        """
        Have BFBuilder use Blub.

        What is Blub : https://esolangs.org/wiki/Blub
        """
        self.increment = "Blub.Blub."
        self.decrement = "Blub!Blub!"
        self.move_right = "Blub.Blub?"
        self.move_left = "Blub?Blub."
        self.input = "Blub.Blub!"
        self.output = "Blub!Blub."
        self.start_loop = "Blub!Blub?"
        self.end_loop = "Blub?Blub!"

    def nyaruko(self):
        """
        Have BFBuilder use Nyaruko lang.

        What is Nyaruko lang : https://github.com/masarakki/nyaruko_lang
        """
        self.increment = "(」・ω・)」うー!(／・ω・)／にゃー!"
        self.decrement = "(」・ω・)」うー!!!(／・ω・)／にゃー!!!"
        self.move_right = "(」・ω・)」うー(／・ω・)／にゃー"
        self.move_left = "(」・ω・)」うー!!(／・ω・)／にゃー!!"
        self.input = "cosmic!"
        self.output = "Let's＼(・ω・)／にゃー"
        self.start_loop = "CHAOS☆CHAOS!"
        self.end_loop = "I WANNA CHAOS!"



from bfi import interpret
from .preference import BFPreference

class BFBuilder:
    """
    Generate Brainf**k source by calling functions.
    """

    def __init__(self):
        self.__text = ""
        self.current = 0

    def generate(self, preference: BFPreference = None) -> None:
        """
        Generate Brainf**k source.

        If you want to generate Brainf**k-like language instead, please give a preference as an argument.
        """
        if preference:
            text = self.__text
            text = text.replace("+", preference.increment)
            text = text.replace("-", preference.decrement)
            text = text.replace(">", preference.move_right)
            text = text.replace("<", preference.move_left)
            text = text.replace(",", preference.input)
            text = text.replace(".", preference.output)
            text = text.replace("[", preference.start_loop)
            text = text.replace("]", preference.end_loop)
            return text
        else:
            return self.__text

    def interpret(self, stdin: str = None, tape_size: int = 30000) -> str:
        """
        Execute Brainf**k source.

        This function uses bfi (https://pypi.org/project/bfi/) internally.
        """
        ret = interpret(self.__text, input_data=stdin, tape_size=tape_size, buffer_output=True)
        return ret

    def input(self) -> None:
        """
        Get user's input and store it to the memory.

        It is equivalent to `,`.
        """
        self.__text += ","

    def output(self) -> None:
        """
        Print a value which is pointed.

        It is equivalent to `.`.
        """
        self.__text += "."

    def move(self, pos) -> None:
        """
        Move to the memory by address.
        """
        if pos - self.current >= 0:
            for _ in range(pos - self.current):
                self.__text += ">"
        else:
            for _ in range(self.current - pos):
                self.__text += "<"
        self.current = pos

    def add(self, num: int) -> None:
        """
        Add a number to the memory.

        You can add a minus number to decrement the value.
        """
        if num > 0:
            for _ in range(num):
                self.__text += "+"
        else:
            for _ in range(-num):
                self.__text += "-"

    def loop(self):
        """
        Generate a loop.

        It is equivalent to `[]`.

        ```
        # Use loop() with `with` statement.
        with builder.loop():
            #
            # Do something here.
            #
        # It will generate `[<expressions>]`.
        ```
        """
        return BFLoop(self)

    def branch(self):
        """
        Generate a branch.

        It looks like the `if` statement in C.

        ```
        # Use branch() with `with` statement.
        with builder.branch():
            #
            # Do something here.
            #
        ```
        """
        return BFBranch(self)

    def init_with_zero(self) -> None:
        """
        Initialize the memory by zero.

        It is equivalent to `[-]`.
        """
        self.__text += "[-]"

    def init_with_num(self, num: int) -> None:
        """
        Initialize the memory by a number given.
        """
        self.init_with_zero()
        self.add(num)

    def init_with_letter(self, letter: str) -> None:
        """
        Initialize the memory by a letter given.
        """
        self.init_with_num(ord(letter))

    def copy(self, dist, tmp) -> None:
        """
        Copy a value to the destination given.

        It requires to use a temporary memory.
        """
        first_pos = self.current
        with self.loop():
            self.add(-1)
            self.move(tmp)
            self.add(1)
        self.move(tmp)
        with self.loop():
            self.add(-1)
            self.move(first_pos)
            self.add(1)
            self.move(dist)
            self.add(1)
        self.move(first_pos)

    def custom(self, text: str) -> None:
        """
        Add custom commands.
        """
        self.__text += text

class BFLoop:
    """
    Express a loop statement on Brainf**k.
    """
    def __init__(self, builder: BFBuilder) -> None:
        self.builder = builder
        self.first_pos = builder.current

    def __enter__(self):
        self.builder.custom("[")
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.builder.move(self.first_pos)
        self.builder.custom("]")

class BFBranch:
    """
    Express a if statement on Brainf**k.
    """
    def __init__(self, builder: BFBuilder) -> None:
        self.builder = builder
        self.first_pos = builder.current

    def __enter__(self):
        self.builder.custom("[[-]")
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.builder.move(self.first_pos)
        self.builder.custom("]")
